<?php

namespace FiloBlu\Flow\Model\From;

/**
 * Class AbstractFrom
 * @package FiloBlu\Flow\Model\From
 */
abstract class AbstractFrom extends \Magento\Framework\Model\AbstractModel
{

    /**
     * @var int
     */
    const PRODUCT_NOT_FOUND_META_STATUS = -2;
    /**
     * @var string
     */
     const XML_PATH_EMAIL_ADDRESS_NOTIFY_ERROR = 'filoblu_flow/flow_email_notification/flow_email_notify_address';
    /**
     * @var string
     */
     const XML_PATH_STORE_NAME = 'general/store_information/name'; 
    
    /**
     * @var string
     */
    protected $_error_message;

    /**
     * Undocumented variable
     *
     * @var array
     */
    protected $_email_notify_errors = [];

    /**
     * @param \Magento\Framework\Model\Context $context
     * @param \Magento\Framework\Registry $registry
     * @param \Magento\Framework\Model\ResourceModel\AbstractResource $resource
     * @param \Magento\Framework\Data\Collection\AbstractDb $resourceCollection
     * @param array $data
     */
    public function __construct(
        \Magento\Framework\Model\Context $context,
        \Magento\Framework\Registry $registry,
        \Magento\Framework\Model\ResourceModel\AbstractResource $resource = null,
        \Magento\Framework\Data\Collection\AbstractDb $resourceCollection = null,
        array $data = []
    ){
        parent::__construct($context, $registry, $resource, $resourceCollection,$data);
    }

    /**
     * @return $this
     */    
    public function setProcessed() {
        $now = date('Y-m-d H:i:s');
        $this->setMetaProcessed(true)
            ->setMetaProcessTime($now);
        return $this;
    }

    /**
     * @param $msg
     */
    public function addEmailNotifyError($msg) {
        $this->_email_notify_errors[] = $msg;
    }

    /**
     * @return array
     */
    public function getEmailNotifyError() {
        return $this->_email_notify_errors;
    }

    /**
     * @param $query
     * @param $file
     * @return mixed
     */
    protected function preprocessFileRows($query,$file) {
        return $query;
    }

    /**
     * @return mixed
     */
    public function getErrorMessage()
    {
        return $this->_error_message;
    }

    /**
     * @param $msg
     */
    public function setErrorMessage($msg)
    {
        $this->_error_message = $msg;
    }

    /**
     * @param $file \FiloBlu\Flow\Model\Inboundflow
     * @return mixed
     * @throws \Exception
     */
    public function processFileRows($file)
    {
        $connection = $this->getResourceCollection()->getConnection();

        $have_executed_on_column = $connection->tableColumnExists($this->getResource()->getMainTable(), 'executed_on');

        // All query to execute by order
        $all_query = [];

        // If I have the 'executed on' column then I'm inside the EAVS
        if ($have_executed_on_column) {
            // 1st
            // All the EAVS actions (meta_ref_id = 0)
            // All the EAVS special actions executed on 'product' (executed_on = 'product')
            $rows = $this->getCollection()
                ->addFieldToFilter('meta_file', $file->getId())
                ->addFieldToFilter('meta_processed', 0)
                ->addFieldToFilter(
                    array('meta_ref_id', 'executed_on'),
                    array(
                        array('eq' => '0'),
                        array('eq' => 'product')
                    )
                );
            $rows->getSelect()->order('e DESC');
            $rows->getSelect()->order('meta_ref_id ASC');
            $all_query[] = $rows;

            // 2nd
            // All the EAVS special actions executed on 'attribute' (executed_on = 'attribute')
            $rows = $this->getCollection()
                ->addFieldToFilter('meta_file', $file->getId())
                ->addFieldToFilter('meta_processed', 0)
                // Greater than 0 are only special actions
                ->addFieldToFilter('meta_ref_id', array('gt' => '0'))
                ->addFieldToFilter('executed_on', 'attribute');
            $rows->getSelect()->order('e DESC');
            $rows->getSelect()->order('meta_ref_id ASC');
            $all_query[] = $rows;

            // 3rd
            // All the EAVS special actions remaining
            $rows = $this->getCollection()
                ->addFieldToFilter('meta_file', $file->getId())
                ->addFieldToFilter('meta_processed', 0)
                // Greater than 0 are only special actions
                ->addFieldToFilter('meta_ref_id', array('gt' => '0'));
            $rows->getSelect()->order('e DESC');
            $rows->getSelect()->order('meta_ref_id ASC');
            $all_query[] = $rows;

            // Fallback to the old default
        } else {
            $rows = $this->getCollection()
                ->addFieldToFilter('meta_file', $file->getId())
                ->addFieldToFilter('meta_processed', 0)
                ->setOrder("meta_ref_id", "ASC");
            $all_query[] = $rows;
        }


        // Running execution in order
        foreach ($all_query as $rows) {
            $query = $connection->query((String)$rows->getSelect());
            $query = $this->preprocessFileRows($query, $file);
            foreach ($query->fetchAll() as $row) {
                try {
                    $this->setData($row);
                    $this->process();
                    $connection->query("Update " . $this->getResource()->getMainTable() . " set meta_processed = 1, meta_process_time = '" . date('Y-m-d H:i:s') . "' where meta_id = " . $row["meta_id"]);
                } catch (\Exception $e) {
                    $this->_logger->error('processing item ' . $row["meta_id"] . ': ' . $e->getMessage());
                   $this->appendErrorMessage('processing item ' . $row["meta_id"] . ': ' . $e->getMessage());
                }
            }
        }

        $this->postprocessFileRows($query);

        $this->postProcess($file);

        // Al termine della fase process dell'import vengono inviate eventuali notifiche di errore via email
        $this->sendErrorNotifications($file);
        
        return $file;

    }

    /**
     * @return mixed
     */
    abstract public function process();

    /**
     * @param $query
     * @return bool
     */
    protected function postprocessFileRows($query)
    {
        return true;
    }

    /**
     * @param \FiloBlu\Flow\Model\Inboundflow $inboundFlow
     * @return bool
     */
    protected function postProcess(\FiloBlu\Flow\Model\Inboundflow $inboundFlow)
    {
        return true;
    }

    /**
     * @param $file
     * @return mixed
     */
    abstract public function sendErrorNotifications($file);

    /**
     * @param $file
     * @return bool
     * @throws \Exception
     */
    public function clean($file)
    {
        if (!$file->getId()) {
            return false;
        }

        $connection = $this->getResource()->getConnection();

        try {
            $sql = "DELETE FROM " . $this->getResource()->getMainTable() . " WHERE meta_file = " . $file->getId();
            $connection->query($sql);
        } catch (\Exception $e) {
            $message = 'processing item ' . $file->getName() . ': ' . $e->getMessage();
            throw new \Exception($message);
        }
    }
    

    /**
     * @param $blocking
     * @param $msg
     * @return bool
     * @throws \Exception
     */
    protected function error($blocking, $msg)
    {
        if ($blocking) {
            throw new \Exception($msg);
        } else {
            $this->setErrorMessage($this->_error_message . "\r\n" . $msg);
            return false;
        }
    }

    /**
     * @param $message
     */
    protected function appendErrorMessage($message){
        $this->_error_message.= PHP_EOL.$message;
    }

}
